"""Extended AST validator."""

# pylint: disable=invalid-name

# import logging

from static_typing.ast_manipulation import AstValidator as AstValidatorBase
import typed_ast.ast3

from .nodes import Comment, Directive

TypedAstValidatorBase = AstValidatorBase[typed_ast.ast3]

# _LOG = logging.getLogger(__name__)


class AstValidator(TypedAstValidatorBase):
    """AST validator for syntax trees obtained by using horast."""

    statement_types = (*TypedAstValidatorBase.statement_types, Comment, Directive)

    expression_types = (*TypedAstValidatorBase.expression_types, Comment)

    def validate_Comment(self, comment):
        assert hasattr(comment, 'comment')
        assert isinstance(comment.comment, str), type(comment.comment)

    def validate_Directive(self, directive):
        assert hasattr(directive, 'expr')
        assert isinstance(directive.expr, str), type(directive.expr)



"""Various helper functions to query and manipulate AST."""

import ast
import logging
import re
import typing as t

import asttokens
import colorama
from static_typing.ast_manipulation import RecursiveAstVisitor
import typed_ast.ast3

from .token_tools import Location, Scope

_LOG = logging.getLogger(__name__)

AstPathNode = t.NamedTuple('AstPathNode', [
    ('node', typed_ast.ast3.AST), ('field', t.Optional[str]), ('index', t.Optional[int]),
    ('scope', Scope)])
"""Node on a AST path.

Meaning of fields:
- node is an AST node on the path
- field is name of the field in that node; None if node is the final (i.e. target) node of the path
- index is index of the next node in the field if value of the field is a list, None otherwise
- scope is location of the above location in the source code
"""


def ast_to_list(
        tree: typed_ast.ast3.AST, only_localizable: bool = False,
        keep_type_ignore: bool = False) -> t.List[typed_ast.ast3.AST]:
    """Generate a flat list of nodes in AST."""
    nodes = []

    class Visitor(RecursiveAstVisitor[typed_ast.ast3]):
        def visit_node(self, node):
            if not keep_type_ignore and isinstance(node, typed_ast.ast3.TypeIgnore):
                return
            if not only_localizable or hasattr(node, 'lineno') and hasattr(node, 'col_offset'):
                nodes.append(node)
    visitor = Visitor()
    visitor.visit(tree)
    return nodes


def get_ast_node_locations(nodes: t.List[typed_ast.ast3.AST]) -> t.List[t.Tuple[int, int]]:
    return [(
        node.lineno if hasattr(node, 'lineno') else None,
        node.col_offset if hasattr(node, 'col_offset') else None) for node in nodes]


def convert_1d_str_index_to_2d(
        text: str, index: int, line_separator: str = '\n',
        newline_starts: t.List[int] = None) -> t.Tuple[int, int]:
    """Convert 1D index in text to tuple (lineno, col_offset)."""
    assert isinstance(text, str), type(text)
    assert isinstance(index, int), type(index)
    if newline_starts is None:
        newline_starts = [m.end() for m in re.finditer(line_separator, text)]
    if index < 0 or index > len(text):
        raise ValueError('index={} is outside [0,{}]'.format(index, len(text)))
    lineno = 1
    for newline_start in newline_starts:
        if newline_start > index:
            break
        lineno += 1
    col_offset = index
    if newline_starts and lineno > 1:
        col_offset -= newline_starts[lineno - 2]
    _LOG.log(logging.NOTSET, 'converted %s[%i] into (%i, %i)',
             repr(text), index, lineno, col_offset)
    return Location(lineno, col_offset)


def get_ast_node_scopes(code: str, nodes: t.List[typed_ast.ast3.AST]) -> t.List[Scope]:
    """Get locations in source code for each given AST node."""
    atok = asttokens.ASTTokens(code, tree=ast.parse(code))
    ast_nodes = ast_to_list(atok.tree)
    assert len(ast_nodes) == len(nodes), (len(ast_nodes), len(nodes))
    newline_starts = [m.end() for m in re.finditer('\n', code)]
    _LOG.debug('new lines in code: %s', newline_starts)
    scopes = []
    _LOG.debug('marked %i nodes:', len(ast_nodes))
    for node in ast_nodes:
        raw_node_scope = atok.get_text_range(node)
        _LOG.debug(
            'node %s at %s: code="""%s%s%s""", tree=%s%s%s',
            type(node).__name__, raw_node_scope, colorama.Style.DIM, atok.get_text(node),
            colorama.Style.RESET_ALL, colorama.Style.DIM, ast.dump(node), colorama.Style.RESET_ALL)
        node_scope = Scope(*[
            convert_1d_str_index_to_2d(code, _, newline_starts=newline_starts)
            for _ in raw_node_scope])
        _LOG.log(logging.NOTSET, 'scope %s is %s', raw_node_scope, node_scope)
        scopes.append(node_scope)
    return scopes


def node_path_in_ast(code: str, tree: typed_ast.ast3.AST, target_node: typed_ast.ast3.AST,
                     *, nodes=None, scopes=None) -> t.List[AstPathNode]:
    """Find path to node in the given AST.

    Return a list of AstPathNode from root node up to the target node.

    Last 2 argument "nodes" and "scopes" are optional but when calling this function many times
    for the same AST providing them is encouraged as constructing those lists is time consuming.
    """
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    assert isinstance(target_node, typed_ast.ast3.AST), type(target_node)
    _LOG.debug('looking for node: %s', typed_ast.ast3.dump(target_node, include_attributes=True))
    if nodes is None:
        nodes = ast_to_list(tree)
    assert target_node in nodes
    if scopes is None:
        scopes = get_ast_node_scopes(code, nodes)
    target_node_index = nodes.index(target_node)
    nodes = nodes[:target_node_index + 1]
    scopes = scopes[:target_node_index + 1]
    node_path = [AstPathNode(target_node, None, None, scopes[target_node_index])]
    current_anchor = target_node
    reversed_anchor_index = 0
    reversed_nodes_scopes = list(reversed(list(enumerate(zip(nodes, scopes)))))
    while current_anchor is not nodes[0]:
        reversed_anchor_index += 1
        for index, (node, scope) in reversed_nodes_scopes[reversed_anchor_index:]:
            _LOG.debug('nodes[%i] is %s', index, node)
            for field_name, field_value in typed_ast.ast3.iter_fields(node):
                if field_value is None or isinstance(field_value, (int, float, str, type, tuple)):
                    continue
                if isinstance(field_value, list):
                    found = False
                    for i, field_value_elem in enumerate(field_value):
                        if field_value_elem is current_anchor:
                            current_anchor = node
                            reversed_anchor_index += index
                            node_path.append(AstPathNode(node, field_name, i, scope))
                            found = True
                            _LOG.debug('"%s[%i]" of %s is on the path', field_name, i, node)
                            break
                    if found:
                        break
                    continue
                if field_value is current_anchor:
                    current_anchor = node
                    reversed_anchor_index += index
                    node_path.append(AstPathNode(node, field_name, None, scope))
                    _LOG.debug('"%s" of %s is on the path', field_name, node)
                    break
    return list(reversed(node_path))


def find_in_ast(code: str, tree: typed_ast.ast3.AST, scope: Scope,
                *, nodes: t.List[typed_ast.ast3.AST] = None,
                scopes=None) -> t.Tuple[t.List[AstPathNode], bool]:
    """Return tuple: (path, before).

    Where:
    - path is path to the anchor node for the target scope
    - before is boolean flag set to True if target scope is before the anchor node, False otherwise

    Last 2 argument "nodes" and "scopes" are optional but when calling this function many times
    for the same AST providing them is encouraged as constructing those lists is time consuming.
    """
    if nodes is None:
        nodes = ast_to_list(tree)
    if scopes is None:
        scopes = get_ast_node_scopes(code, nodes)
    node_scopes_by_start = list(zip(nodes, scopes))
    node_scopes_by_start.sort(key=lambda _: _[1].end, reverse=True)
    node_scopes_by_start.sort(key=lambda _: _[1].start)
    _LOG.debug('by start: %s', node_scopes_by_start)
    node_scopes_by_end = list(zip(nodes, scopes))
    node_scopes_by_end.sort(key=lambda _: _[1].start, reverse=True)
    node_scopes_by_end.sort(key=lambda _: _[1].end)
    _LOG.debug('by end: %s', node_scopes_by_end)
    target_scope = scope
    _LOG.debug('the target scope %s', target_scope)

    node_by_end_after_index = None
    for i, (node, scope_) in enumerate(node_scopes_by_end):
        if target_scope.start > scope_.end:
            node_by_end_after_index = i
            _LOG.debug('is after node %s at %s', node, scope_)
            break
    if node_by_end_after_index is None:
        _LOG.debug('is not after any node')

    node_by_start_before_index = None
    for i, (node, scope_) in enumerate(node_scopes_by_start):
        if target_scope.end < scope_.start:
            node_by_start_before_index = i
            _LOG.debug('is before node %s at %s', node, scope_)
            break
    if node_by_start_before_index is None:
        _LOG.debug('is not before any node')

    scopes_containing_target_scope = []
    for i, (node, scope_) in enumerate(node_scopes_by_start):
        if scope_.start > target_scope.start:
            break
        if scope_.end > target_scope.end:
            scopes_containing_target_scope.append((node, scope_))
    _LOG.debug('is within nodes %s', scopes_containing_target_scope)

    if node_by_end_after_index is None:
        _LOG.debug('target %s is before first node', target_scope)
        assert isinstance(nodes[0], (typed_ast.ast3.Module, typed_ast.ast3.Interactive)), \
            type(nodes[0])
        return ([AstPathNode(nodes[0], 'body', 0, scopes[0])], True)

    if node_by_start_before_index is None and not scopes_containing_target_scope:
        _LOG.debug('target %s is after last node', target_scope)
        assert isinstance(nodes[0], (typed_ast.ast3.Module, typed_ast.ast3.Interactive)), \
            type(nodes[0])
        assert len(nodes[0].body) > 0
        return ([AstPathNode(nodes[0], 'body', len(nodes[0].body) - 1, scopes[0])], False)

    if node_by_start_before_index is None or not scopes_containing_target_scope:
        raise NotImplementedError(
            'inconsistent results for target {} in:\n"""\n{}\nafter {}, before {}, within {}"""'
            .format(target_scope, code, node_by_end_after_index, node_by_start_before_index,
                    scopes_containing_target_scope))

    assert node_by_end_after_index is not None
    assert node_by_start_before_index is not None
    assert scopes_containing_target_scope
    _LOG.debug(
        'target %s is neither before first node nor after last node in:\n"""\n%s\n"""'
        '\nbut between %s and %s, within %s',
        target_scope, code, node_scopes_by_end[node_by_end_after_index],
        node_scopes_by_start[node_by_start_before_index], scopes_containing_target_scope)
    within_node, _ = scopes_containing_target_scope[-1]
    before_node, _ = node_scopes_by_start[node_by_start_before_index]
    path = node_path_in_ast(code, tree, before_node, nodes=nodes, scopes=scopes)
    assert len(path) >= 2, path
    assert path[-2].node is within_node, (path[-2].node, within_node)
    return (path[:-1], True)


def insert_at_path_in_tree(
        tree: typed_ast.ast3.AST, inserted: typed_ast.ast3.AST,
        path_to_anchor: t.Sequence[AstPathNode],
        before_anchor: bool = False) -> typed_ast.ast3.AST:
    """Insert a new AST node into an existing AST at exactly specified location."""
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    assert isinstance(inserted, typed_ast.ast3.AST), type(inserted)
    # assert isinstance(anchor, typed_ast.ast3.AST), type(anchor)
    parent, field, index = path_to_anchor[-1][:3]
    if not before_anchor:
        index += 1
    getattr(parent, field).insert(index, inserted)
    return tree


def insert_in_tree(
        tree: typed_ast.ast3.AST, inserted: typed_ast.ast3.AST, anchor: typed_ast.ast3.AST,
        before_anchor: bool = False, strict: bool = False) -> typed_ast.ast3.AST:
    """Deprecated, use insert_at_path_in_tree instead.

    Insert a new AST node into an existing AST near the anchor node.

    Try to maintain correctness after insertion.
    """
    assert isinstance(tree, typed_ast.ast3.AST), type(tree)
    assert isinstance(inserted, typed_ast.ast3.AST), type(inserted)
    assert isinstance(anchor, typed_ast.ast3.AST), type(anchor)
    node_path = node_path_in_ast('', tree, anchor)
    if node_path is None:
        raise ValueError('the anchor node {} not found in AST {}'.format(anchor, tree))
    _LOG.debug('node path: %s', node_path)
    try:
        parent, field, index = node_path[-2]
    except IndexError as err:
        if strict:
            raise NotImplementedError('cannot insert at AST root') from err
        parent, field, index = node_path[0].node, 'body', 0
    _LOG.debug('parent of anchor node: %s', parent)
    if strict and index is None:
        raise NotImplementedError('cannot insert into a non-list field')
    node_path_index = -2
    if field == 'targets':
        index = None  # don't insert anything into assignment
    while index is None:
        node_path_index -= 1
        parent, field, index = node_path[node_path_index]
        if field == 'targets':
            index = None  # don't insert anything into assignment
    if before_anchor:
        pass
    else:
        index += 1
    getattr(parent, field).insert(index, inserted)
    return tree



"""Default logging mechanism for horast module."""

import logging
import logging.config
import os
import pathlib
import platform
import typing as t

import colorama

LOGS_PATHS = {
    'Linux': pathlib.Path('~', '.local', 'share'),
    'Darwin': pathlib.Path('~', 'Library', 'Logs'),
    'Windows': pathlib.Path('%LOCALAPPDATA%')}

LOGS_PATH = LOGS_PATHS[platform.system()].joinpath('horast')


def normalize_path(path: t.Union[pathlib.Path, str]) -> t.Union[pathlib.Path, str]:
    if isinstance(path, str):
        return os.path.expanduser(os.path.expandvars(path))
    return pathlib.Path(normalize_path(str(path)))


def logging_level_from_envvar(envvar: str, default: int = logging.WARNING) -> int:
    """Translate text envvar value into an integer corresponding to a logging level."""
    envvar_value = os.environ.get(envvar)
    if envvar_value is None:
        return default
    envvar_value = envvar_value.upper()
    if not hasattr(logging, envvar_value):
        try:
            return int(envvar_value)
        except ValueError:
            return default
    return getattr(logging, envvar_value)


colorama.init()

if not normalize_path(LOGS_PATH).is_dir():
    normalize_path(LOGS_PATH).mkdir(parents=True)

logging.config.dictConfig({
    'formatters': {
        'brief': {
            '()': 'colorlog.ColoredFormatter',
            'style': '{',
            'format': '{name} [{log_color}{levelname}{reset}] {message}'},
        'precise': {
            'style': '{',
            'format': '{asctime} {name} [{levelname}] {message}'}},
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'brief',
            'level': logging_level_from_envvar('LOGGING_LEVEL'),
            'stream': 'ext://sys.stdout'},
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'precise',
            'level': logging.NOTSET,
            'filename': normalize_path(str(LOGS_PATH.joinpath('horast.log'))),
            'maxBytes': 1 * 1024 * 1024,
            'backupCount': 10}},
    'root': {
        'handlers': ['console', 'file'],
        'level': logging.NOTSET},
    'version': 1,
    'disable_existing_loggers': False})



"""Extract comments from untouched source code and insert them into the AST."""

import logging
import tokenize
import typing as t
import warnings

import typed_ast.ast3
import typed_astunparse

from .nodes import Comment, Directive, Pragma, OpenMpPragma, OpenAccPragma, Include
from .token_tools import get_token_scope, get_token_locations  # , get_token_scopes
from .ast_tools import \
    ast_to_list, get_ast_node_locations, get_ast_node_scopes, find_in_ast, \
    insert_at_path_in_tree, insert_in_tree

_LOG = logging.getLogger(__name__)

CLASSIFIED_NODES = (Include, OpenMpPragma, OpenAccPragma, Pragma, Directive)


def is_prefixed(text: str, prefix: str) -> bool:
    """Check if a text (assumed to be a token value) is prefixed with a given prefix.

    This is different from simple checking text.startswith(prefix),
    because it also applies criteria normally applied by tokenizer to separate tokens.

    E.g. "acc loop" is prefixed with "acc", but "accelerate" is not.
    """
    if not text.startswith(prefix):
        return False
    if len(text) == len(prefix) or prefix[-1] == ':':
        return True
    return any(text[len(prefix):].startswith(_) for _ in (' ', '('))


def classify_comment_token(token: tokenize.TokenInfo) -> type:
    # for node_type, prefixes in PREFIXES:
    for node_type in CLASSIFIED_NODES:
        # for prefix in prefixes:
        for prefix in getattr(node_type, '_comment_prefixes', ()):
            if is_prefixed(token.string[1:], prefix):
                _LOG.debug('classified "%s" as %s (prefix: %s)', token.string, node_type, prefix)
                return node_type
    _LOG.debug('classified "%s" as %s', token.string, Comment)
    return Comment


def insert_comment_token(token: tokenize.TokenInfo, code: str, tree: typed_ast.ast3.AST,
                         *, nodes=None, scopes=None):
    """Insert comment into the syntax tree.

    Last 2 argument "nodes" and "scopes" are optional but when calling this function many times
    for the same AST providing them is encouraged as constructing those lists is time consuming.
    """
    if nodes is None:
        nodes = ast_to_list(tree)
    if scopes is None:
        scopes = get_ast_node_scopes(code, nodes)
    scope = get_token_scope(token)
    path_to_anchor, before_anchor = find_in_ast(code, tree, scope, nodes=nodes, scopes=scopes)
    node_type = classify_comment_token(token)
    if issubclass(node_type, Comment):
        node = node_type.from_token(token, path_to_anchor, before_anchor)
    else:
        if issubclass(node_type, Directive):
            node = node_type.from_token(token)
        else:
            raise ValueError('insertion of node {} (from token "{}") is not supported'
                             .format(node_type.__name__, token))

    _LOG.debug('inserting a %s: %s %s %s', type(node).__name__, node,
               'before' if before_anchor else 'after', path_to_anchor[-1])
    return insert_at_path_in_tree(tree, node, path_to_anchor, before_anchor)


def insert_comment_tokens(
        code: str, tree: typed_ast.ast3.AST,
        tokens: t.List[tokenize.TokenInfo]) -> typed_ast.ast3.AST:
    """Insert comment tokens into an AST obtained from typed_ast parser."""
    assert isinstance(tree, typed_ast.ast3.AST)
    assert isinstance(tokens, list)
    nodes = ast_to_list(tree)
    scopes = get_ast_node_scopes(code, nodes)
    assert len(nodes) == len(scopes), (len(nodes), len(scopes))
    for token in tokens:
        tree = insert_comment_token(token, code, tree, nodes=nodes, scopes=scopes)
    return tree


def insert_comment_tokens_approx(
        tree: typed_ast.ast3.AST, tokens: t.List[tokenize.TokenInfo]) -> typed_ast.ast3.AST:
    """Deprecated, use insert_comment_tokens instead."""
    warnings.warn('function insert_comment_tokens_approx is outdated and faulty, and it will be'
                  ' removed from horast, use insert_comment_tokens instead', DeprecationWarning)
    assert isinstance(tree, typed_ast.ast3.AST)
    assert isinstance(tokens, list)
    token_locations = get_token_locations(tokens)
    _LOG.debug('token locations: %s', token_locations)
    nodes = ast_to_list(tree, only_localizable=True)
    if not nodes and tokens:
        _LOG.debug('overwriting empty AST with simplest editable tree')
        tree = typed_ast.ast3.Module(body=[], type_ignores=[], lineno=1, col_offset=0)
        nodes = ast_to_list(tree, only_localizable=True)
    node_locations = get_ast_node_locations(nodes)
    _LOG.debug('node locations: %s', node_locations)
    node_locations_iter = enumerate(node_locations)
    token_insertion_indices = []
    tokens_eol_status = []
    for token_index, token_location in enumerate(token_locations):
        eol_comment_here = False
        try:
            node_index, node_location = next(node_locations_iter)
        except StopIteration:
            node_index = len(node_locations)
            node_location = None
        while node_location is not None:
            token_line, _ = token_location
            node_line, _ = node_location
            if node_line > token_line:
                break
            if node_line == token_line:
                eol_comment_here = True
                if node_index < len(node_locations) - 1:
                    next_node_line, _ = node_locations[node_index + 1]
                    if next_node_line == token_line:
                        eol_comment_here = False
                # if eol_comment_here:
                #    raise NotImplementedError(
                #        'code "{}" and comment "{}" in line {}'
                #        ' -- only whole line comments are currently supported'
                #        .format(typed_astunparse.unparse(nodes[node_index]).strip(),
                #                tokens[token_index].string, node_line))
            try:
                node_index, node_location = next(node_locations_iter)
            except StopIteration:
                node_index = len(node_locations)
                break
        tokens_eol_status.append(eol_comment_here)
        token_insertion_indices.append(node_index)
    _LOG.debug('token insertion indices: %s', token_insertion_indices)
    _LOG.debug('tree before insertion:\n"""\n%s\n"""', typed_astunparse.dump(tree))
    _LOG.debug('code before insertion:\n"""\n%s\n"""', typed_astunparse.unparse(tree).strip())
    for token_index, token_insertion_index in reversed(list(enumerate(token_insertion_indices))):
        token = tokens[token_index]
        eol = tokens_eol_status[token_index]
        comment = Comment(token.string[1:], eol)
        if token_insertion_index == 0:
            anchor = nodes[token_insertion_index]
            before_anchor = True
        elif token_insertion_index == len(node_locations):
            anchor = nodes[-1]
            before_anchor = False
        else:
            anchor = nodes[token_insertion_index - 1]
            before_anchor = False
        _LOG.debug('inserting %s %s %s', comment, 'before' if before_anchor else 'after', anchor)
        tree = insert_in_tree(tree, comment, anchor=anchor, before_anchor=before_anchor)
    _LOG.debug('tree after insertion:\n"""\n%s\n"""', typed_astunparse.dump(tree))
    # _LOG.warning('code after insertion:\n"""\n%s\n"""', typed_astunparse.unparse(tree).strip())
    return tree



"""Various helper functions to query and manipulate tokens."""

import io
import tokenize
import typing as t
import warnings

Location = t.NamedTuple('Location', [('lineno', int), ('offset', int)])

Scope = t.NamedTuple('Scope', [('start', t.Tuple[int, int]), ('end', t.Tuple[int, int])])


def get_token_scope(token: tokenize.TokenInfo):
    return Scope(Location(*token.start), Location(*token.end))


def get_tokens(code: str, token_filter=None) -> t.List[tokenize.TokenInfo]:
    """List of all tokens contained in the given code."""
    assert isinstance(code, str), type(code)
    code_bytes = code.encode()  # type: bytes
    with io.BytesIO(code_bytes) as code_bytes_reader:
        tokenizer = tokenize.tokenize(code_bytes_reader.readline)
        tokens = [token for token in tokenizer if token_filter is None or token_filter(token)]
    return tokens


def is_comment(token: tokenize.TokenInfo) -> bool:
    return token.type is tokenize.COMMENT


def is_type_comment(token: tokenize.TokenInfo) -> bool:
    return token.string.startswith('# type:') and token.string != '# type: ignore'


def is_comment_but_not_type_comment(token: tokenize.TokenInfo) -> bool:
    return is_comment(token) and not is_type_comment(token)


def get_comment_tokens(code: str, ignore_type_comments: bool = True) -> t.List[tokenize.TokenInfo]:
    return get_tokens(code, is_comment_but_not_type_comment if ignore_type_comments else is_comment)


def get_token_locations(tokens: t.List[tokenize.TokenInfo]) -> t.List[t.Tuple[int, int]]:
    warnings.warn('function get_token_locations is obsolete and it will be removed from horast,'
                  ' use get_token_scope instead', DeprecationWarning)
    return [token.start for token in tokens]



"""horast: Human-oriented abstract syntax tree parser and unparser."""

from static_typing import dump

from .parser import parse
from .unparser import unparse

from .ast_validator import AstValidator

__all__ = ['dump', 'parse', 'unparse', 'AstValidator']



"""Definitions of AST nodes used to store missing syntax information."""

# pylint: disable=too-few-public-methods

import logging
import tokenize
import typing as t

import typed_ast.ast3

from .token_tools import get_token_scope

_LOG = logging.getLogger(__name__)


class Comment(typed_ast.ast3.AST):
    """Store code comment in AST.

    Examples:

    # full line comment

    expr # end-of-line comment

    (value1,
     # full line comment inside an expression
     value2,
     value3)

    (value1,
     value2,  # end-of line comment inside an expression
     value3)
    """

    _fields = typed_ast.ast3.AST._fields + ('comment', 'eol')

    @staticmethod
    def is_eol(token: tokenize.TokenInfo, path_to_anchor, before_anchor) -> bool:
        """Determine if a given comment token is an end-of-line comment or not."""
        if before_anchor:
            return False
        assert path_to_anchor
        anchor = path_to_anchor[-1]
        assert isinstance(anchor.field, str), type(anchor.field)
        node = getattr(anchor.node, anchor.field)
        if anchor.index is not None:
            node = node[anchor.index]
        if not hasattr(node, 'lineno'):
            raise ValueError('anchor node {} must have "lineno" attribute'
                             .format(typed_ast.ast3.dump(node, include_attributes=True)))
        token_scope = get_token_scope(token)
        assert token_scope.start.lineno == token_scope.end.lineno, token_scope
        assert token_scope.start.offset < token_scope.end.offset, token_scope
        eol = anchor.scope.end.lineno == token_scope.start.lineno
        _LOG.debug('comment scope: %s, anchor scope: %s, evaluated EOL status: %s',
                   (token.start[0], token.end[0], token_scope), (node.lineno, anchor.scope), eol)
        return eol

    @classmethod
    def from_token(cls, token: tokenize.TokenInfo, path_to_anchor, before_anchor):
        eol = cls.is_eol(token, path_to_anchor, before_anchor)
        return cls(
            comment=token.string[1:], eol=eol,
            lineno=token.start[0], col_offset=token.start[1])


class BlockComment(typed_ast.ast3.AST):
    """Store sequence of code comments as a single node in AST.

    Example:

    # 1st line of a comment
    # 2nd line of a comment
    # 3rd line of a comment
    """

    _fields = typed_ast.ast3.AST._fields + ('comments',)

    @classmethod
    def from_token(cls, token: tokenize.TokenInfo, *extra_tokens: t.List[tokenize.TokenInfo]):
        raise NotImplementedError()


class Directive(typed_ast.ast3.AST):
    """Store directive in AST.

    In Python, directives would be expressed as comments, but may have special additional meaning.

    Examples:

    #if
    #else
    #endif
    #def
    #undef
    #ifdef
    #ifndef
    """

    _comment_prefixes = ('if', 'else', 'endif', 'def', 'undef', 'ifdef', 'ifndef')

    _fields = typed_ast.ast3.AST._fields + ('expr',)

    @classmethod
    def from_token(cls, token: tokenize.TokenInfo):
        """Create a Directive from a comment token."""
        expr = token.string[1:]
        if len(cls._comment_prefixes) == 1:
            prefix = cls._comment_prefixes[0]
            assert expr.startswith(prefix), expr
            _LOG.debug(
                'stripping prefix "%s" and following whitespace from the token %s', prefix, token)
            expr = expr[len(prefix):].lstrip()
        return cls(
            expr=expr,
            lineno=token.start[0], col_offset=token.start[1])


class Pragma(Directive):
    """Store a pragma in AST.

    Examples:

    # pragma: once
    # pragma: ...
    """

    _comment_prefixes = (' pragma:',)


class OpenMpPragma(Pragma):
    """A special node for storing OpenMP pragmas in AST.

    Examples:

    # pragma: omp parallel for
    """

    _comment_prefixes = (' pragma: omp',)


class OpenAccPragma(Pragma):
    """A special node for storing OpenACC pragmas in AST.

    Examples:

    # pragma: acc parallel
    """

    _comment_prefixes = (' pragma: acc',)


class Include(Directive):
    """Store an include directive in AST.

    # include: <cstdio>
    # include: "mpif.h"
    # include: my_header.h
    """

    _comment_prefixes = (' include:',)
